import re
import operator
import random

import pygame

import data
from world import *


TARGETS = []

def EVERYTHING(objects, m=None):
    return objects
TARGETS.append((EVERYTHING, r"""(EVERYTHING|ANYTHING)"""))

def WHOEVER(objects, m=None):
    for object in objects:
        if isinstance(object, Person):
            yield object
TARGETS.append((WHOEVER, r"""(WHOEVER IS|EVERYBODY|ANYONE|PLAYERS|EVERYONE)"""))

def ALL_STAR(objects, m=None):
    what = m.groupdict()['WHAT'].rstrip('S')
    for object in objects:
        if re.match(object.what, what):
            yield object
TARGETS.append((ALL_STAR, r"""ALL (?P<WHAT>\w+)"""))


FILTERS = []

def compare_distance(compare, distance, allow_self=False):
    def filter_(objects, world, groups, m=None):
        for object in objects:
            for item in world:
                if re.match(item.what, m.groupdict()['OBJECT']):
                    dx = abs(object.rect.x - item.rect.x)
                    dy = abs(object.rect.y - item.rect.y)
                    if (allow_self or (object != item)) and (
                            compare(dx, distance) and
                            compare(dy, distance)):
                        yield object
    return filter_

CLOSE_TO_STAR = compare_distance(operator.lt, 64)
FILTERS.append((CLOSE_TO_STAR, r"""CLOSE TO (?P<OBJECT>\w+)"""))

TOUCHING_STAR = compare_distance(operator.lt, 32)
FILTERS.append((TOUCHING_STAR, r"""TOUCHING (?P<OBJECT>\w+)"""))

OVER_STAR = compare_distance(operator.lt, 16)
FILTERS.append((OVER_STAR, r"""OVER (?P<OBJECT>\w+)"""))

FAR_FROM_STAR = compare_distance(operator.gt, 64, allow_self=True)
FILTERS.append((FAR_FROM_STAR, r"""FAR FROM (?P<OBJECT>\w+)"""))

def NAMED(objects, world, groups, m=None):
    for object in objects:
        name = getattr(object, 'name', 'Nameless')
        if re.match(name, m.groupdict()['NAME']):
            yield object
FILTERS.append((NAMED, r"""(NAMED|CALLED) (?P<NAME>\w+)"""))


ACTIONS = []

def WINS(objects, world, groups, m=None):
    for object in objects:
        object.winner = True
ACTIONS.append((WINS, r"""WINS?"""))

def MOVE_DIRECTION(objects, world, groups, m=None):
    for object in objects:
        direction = {'NORTH'    : ( 0, -1), 'N' : ( 0, -1),
                     'SOUTH'    : ( 0,  1), 'S' : ( 0,  1),
                     'EAST'     : ( 1,  0), 'E' : ( 1,  0),
                     'WEST'     : (-1,  0), 'W' : (-1,  0),
                     'NORTHWEST': (-1, -1), 'NW': (-1, -1),
                     'NORTHEAST': ( 1, -1), 'NE': ( 1, -1),
                     'SOUTHWEST': (-1,  1), 'SW': (-1,  1),
                     'SOUTHEAST': ( 1,  1), 'SE': ( 1,  1),
                    }[m.groupdict()['DIRECTION']]
        object.direction = direction

        # Add speed. 
        object._init_queue.insert(0,
                lambda s: setattr(s, 'speed', 3))
        object._finish_queue.insert(0,
                lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((MOVE_DIRECTION, r"""MOVE (?P<DIRECTION>NORTHWEST|NORTHEAST|NORTH|SOUTHWEST|SOUTHEAST|SOUTH|EAST|WEST|NW|NE|N|SW|SE|S|W|E)"""))

def MOVE(objects, world, groups, m=None):
    for object in objects:
        speed = m.groupdict()['SPEED']
        speed = {'FAST': 8,
                 'SUPER FAST': 12,
                 'SLOW': 2,
                 'SUPER SLOW': 1,
                 None: 3}[speed]

        # Add speed. 
        object._init_queue.insert(0,
                lambda s, speed=speed: setattr(s, 'speed', speed))

        # Add random direction only to non-players.
        if not isinstance(object, FreeWilling):
            direction = (random.choice([-1,0,1]), random.choice([-1,0,1]))
            object._init_queue.insert(0,
                    lambda s, direction=direction: setattr(s, 'direction', direction))
            
        object._finish_queue.insert(0,
                lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((MOVE, r"""MOVES?(\s(?P<SPEED>(SUPER )?(FAST|SLOW)))?"""))

def FLY(objects, world, groups, m=None):
    for object in objects:
        # Remove all collidables from this object, so it doesn't hit anything.
        object._init_queue.insert(0, 
                lambda s: setattr(s, 'collidables', {}))
        # Also, remove object from its group, so nothing collides with it.
        object._init_queue.insert(0,
                lambda s: s.remove(groups[s.group]))
        # And restore everything.
        object._finish_queue.insert(0, 
                lambda s, c=object.collidables: setattr(s, 'collidables', c))
        object._finish_queue.insert(0,
                lambda s: s.add(groups[s.group]))
ACTIONS.append((FLY, r"""(FLIES|FLY)"""))

def WALK_OVER_WATER(objects, world, groups, m=None):
    for object in objects:
        copy_ = object.collidables.copy()
        copy_['WATER'] = 0.5
        object._init_queue.insert(0,
                lambda s: setattr(s, 'collidables', copy_))
        object._finish_queue.insert(0,
                lambda s, c=object.collidables: setattr(s, 'collidables', c))
ACTIONS.append((WALK_OVER_WATER, r"""(WALKS? OVER WATER|CAN SWIM|SWIMS)"""))

def REPRODUCE(objects, world, groups, m=None):
    for object in objects:
        if random.random() < 0.001:
            new_ = object.copy()
            width = new_.rect.width
            height = new_.rect.height
            new_.rect.x = random.choice([new_.rect.x - width, new_.rect.x + width])
            new_.rect.y = random.choice([new_.rect.y - height, new_.rect.y + height])
            if object.group in groups:
                groups[object.group].add(new_)
            index = groups['ALL'].spritelist.index(object)
            groups['ALL'].add(new_, index)  # add to allsprites
            world.append(new_)
ACTIONS.append((REPRODUCE, r"""(REPRODUCE|REPLICATE)"""))

def INVISIBLE(objects, world, groups, m=None):
    for object in objects:
        pixel = pygame.image.load(data.load("invisible_1x1.png")).convert_alpha()
        image = pygame.transform.scale(pixel, object.image.get_size())
        object._init_queue.insert(0,
                lambda s: setattr(s, 'image', image))
        object._init_queue.insert(0, 
                lambda s: setattr(s, '_images', []))
        object._finish_queue.insert(0,
                lambda s, image=object.image: setattr(s, 'image', image))
        object._finish_queue.insert(0,
                lambda s, images=object._images: setattr(s, '_images', images))
ACTIONS.append((INVISIBLE, r"""(IS|ARE) INVISIBLE"""))

def DIE(objects, world, groups, m=None):
    for object in objects:
        object.alive = False
        object.kill()
ACTIONS.append((DIE, r"""(DIE|DIES|IS DEAD|ARE DEAD|DISAPPEAR|DISAPPEARS)"""))

def TELEPORT(objects, world, groups, m=None):
    for object in objects:
        object.rect.x = random.randint(0, 640-32)
        object.rect.y = random.randint(32, 480-64)

        # Remove from allsprites to check collision,
        # but first get the index for ordering.
        index = groups['ALL'].spritelist.index(object)
        groups['ALL'].remove(object)
        while pygame.sprite.spritecollide(object, groups['ALL'], 0):
            object.rect.x = random.randint(0, 640-32)
            object.rect.y = random.randint(32, 480-64)
        groups['ALL'].add(object, index)  # add to allsprites
ACTIONS.append((TELEPORT, r"""(TELEPORT|APPARATE)"""))

def this_to_that(groups, what, group, image):
    def func(obj, image=image):
        obj.remove(groups[obj.group])
        obj.what = what
        obj.group = group
        if isinstance(image, basestring):
            image = pygame.image.load(data.load(image)).convert_alpha()
        obj.image = image
        obj.add(groups.setdefault(group, pygame.sprite.Group()))
    return func

def TRANSFORM_INTO_STAR(objects, world, groups, m=None):
    for object in objects:
        target = globals()[m.groupdict()['TARGET'].title()]
        index = groups['ALL'].spritelist.index(object)
        object.kill()
        new_ = target((object.rect.x, object.rect.y), object.speed, world)
        i = world.index(object)
        del world[i]
        world.append(new_)
        new_.add(groups.setdefault(new_.group, pygame.sprite.Group()))
        groups['ALL'].add(new_, index)
ACTIONS.append((TRANSFORM_INTO_STAR, r"""(TRANSFORMS? (IN|TO|INTO)|BECOMES?) (?P<TARGET>\w+)"""))

def FREEZE(objects, world, groups, m=None):
    for object in objects:
        if re.match(object.what, 'WATER'):
            object._init_queue.insert(0, this_to_that(groups, 'ICE', 'ICE', 'ice.png'))
            object._finish_queue.insert(0, this_to_that(groups, '(WATER|LAKE|POND)', 'WATER', object.image))
        else:
            object._init_queue.insert(0,
                    lambda s: setattr(s, 'speed', 0))
            object._finish_queue.insert(0,
                    lambda s, speed=object.speed: setattr(s, 'speed', speed))
ACTIONS.append((FREEZE, r"""FREEZES?"""))

def MUTE(objects, world, groups, m=None):
    for object in objects:
        object._init_queue.insert(0,
                lambda s: setattr(s, 'mute', True))
        object._finish_queue.insert(0,
                lambda s: setattr(s, 'mute', False))
ACTIONS.append((MUTE, r"""(IS|ARE) MUTE"""))


def check_rule(rule):
    for select, regexp in TARGETS:
        m = re.match(regexp, rule)
        if m:
            rule = rule[len(m.group()):].strip()
            break
    else:
        return False

    for filter_, regexp in FILTERS:
        m = re.match(regexp, rule)
        if m:
            rule = rule[len(m.group()):].strip()
            break

    for action, regexp in ACTIONS:
        m = re.match(regexp, rule)
        if m:
            return True


def apply_rule(rule, objects, groups):
    for select, regexp in TARGETS:
        m = re.match(regexp, rule)
        if m:
            targets = select(objects, m)
            rule = rule[len(m.group()):].strip()
            break

    for filter_, regexp in FILTERS:
        m = re.match(regexp, rule)
        if m:
            targets = filter_(targets, objects, groups, m)
            rule = rule[len(m.group()):].strip()
            break

    for action, regexp in ACTIONS:
        m = re.match(regexp, rule)
        if m:
            action(targets, objects, groups, m)

